"""A Pure-Python Twisted library for LDAP"""
__version__ = "20.1.1"

__title__ = "ldaptor"
__description__ = "A Pure-Python Twisted library for LDAP"
__uri__ = "https://github.com/twisted/ldaptor"

__license__ = "MIT"
__author__ = "The ldaptor developers"
__copyright__ = "Copyright (c) 2002-2019 {}".format(__author__)
